@@ -1,9 +1,91 @@
 from __future__ import print_function
 import multiprocessing
 import warnings
+import dill
 from tqdm.auto import tqdm
-from dill import dumps, loads
 from traceback import format_exc
+from pickle import PicklingError
+
+from pickle import _Pickler, dispatch_table
+
+
+class Pickler(_Pickler):
+    def save(self, obj, save_persistent_id=True):
+        self.framer.commit_frame()
+
+        # Check for persistent id (defined by a subclass)
+        pid = self.persistent_id(obj)
+        if pid is not None and save_persistent_id:
+            self.save_pers(pid)
+            return
+
+        # Check the memo
+        x = self.memo.get(id(obj))
+        if x is not None:
+            self.write(self.get(x[0]))
+            return
+
+        rv = NotImplemented
+        reduce = getattr(self, "reducer_override", None)
+        if reduce is not None:
+            rv = reduce(obj)
+
+        if rv is NotImplemented:
+            # Check the type dispatch table
+            t = type(obj)
+            f = self.dispatch.get(t)
+            if f is not None:
+                f(self, obj)  # Call unbound method with explicit self
+                return
+
+            # Check private dispatch table if any, or else
+            # copyreg.dispatch_table
+            reduce = getattr(self, 'dispatch_table', dispatch_table).get(t)
+            if reduce is not None:
+                rv = reduce(obj)
+            else:
+                # Check for a class with a custom metaclass; treat as regular
+                # class
+                if issubclass(t, type):
+                    self.save_global(obj)
+                    return
+
+                # Check for a __reduce_ex__ method, fall back to __reduce__
+                reduce = getattr(obj, "__reduce_ex__", None)
+                try:
+                    if reduce is not None:
+                        rv = reduce(self.proto)
+                    else:
+                        reduce = getattr(obj, "__reduce__", None)
+                        if reduce is not None:
+                            rv = reduce()
+                        else:
+                            raise PicklingError("Can't pickle %r object: %r" %
+                                                (t.__name__, obj))
+                except:
+                    rv = (lambda *args, **kwargs: None, ())
+
+        # Check for string returned by reduce(), meaning "save as global"
+        if isinstance(rv, str):
+            self.save_global(obj, rv)
+            return
+
+        # Assert that reduce() returned a tuple
+        if not isinstance(rv, tuple):
+            raise PicklingError("%s must return string or tuple" % reduce)
+
+        # Assert that it returned an appropriately sized tuple
+        l = len(rv)
+        if not (2 <= l <= 6):
+            raise PicklingError("Tuple returned by %s must have "
+                                "two to six elements" % reduce)
+
+        # Save the reduce() output and finally memoize the object
+        self.save_reduce(obj=obj, *rv)
+
+dill.StockPickler = Pickler
+
+
 
 def chunks(n, *args):
     """ Yield successive n-sized chunks from lists. """
@@ -17,9 +99,7 @@ def chunks(n, *args):
             yield [a[i:i+n] for a in args]
 
 class tqdmm(tqdm):
-    """ Overload tqdm to make a special version of tqdm functioning as a meter
-        wp@tl20200204
-    """
+    """ Overload tqdm to make a special version of tqdm functioning as a meter. """
 
     def __init__(self, *args, **kwargs):
         self._n = 0
@@ -47,18 +127,20 @@ def parfor(*args, **kwargs):
     """ @parfor(iterator=None, args=(), kwargs={}, length=None, desc=None, bar=True, qbar=True, np1/3 serial=4, debug=False):
         decorator to parallize for-loops
 
-        required inputs:
+        required arguments:
             fun:    function taking arguments: iteration from  iterable, other arguments defined in args & kwargs
             iterable: iterable from which an item is given to fun as a first argument
 
-        optional inputs:
+        optional arguments:
             args:   tuple with other unnamed arguments to fun
             kwargs: dict with other named arguments to fun
+            length: give the length of the iterator in cases where len(iterator) results in an error
             desc:   string with description of the progress bar
             bar:    bool enable progress bar
             pbar:   bool enable buffer indicator bar
             nP:     number of workers, default: number of cpu's/3
             serial: switch to serial if number of tasks less than serial, default: 4
+            debug:  if an error occurs in an iteration, return the erorr instead of retrying in the main process
 
         output:       list with results from applying the decorated function to each iteration of the iterator
                       specified as the first argument to the function
@@ -120,11 +202,6 @@ def parfor(*args, **kwargs):
             fun
 
             >> [0, 3, 12, 27, 48, 75, 108, 147, 192, 243]
-
-        wp@tl20190816 -- try to use bare multiprocessing
-        wp@tl20200121 -- use deco
-        wp@tl20200204 -- finally figure out how to use multiprocessing
-        wp@tl20200831 -- package
         """
     def decfun(fun):
         return pmap(fun, *args, **kwargs)
@@ -171,7 +248,7 @@ class parpool(object):
 
     @fun.setter
     def fun(self, fun):
-        funs = dumps(fun, recurse=True)
+        funs = dill.dumps(fun, recurse=True)
         self._fun = (fun, hash(funs), funs)
 
     @property
@@ -180,7 +257,7 @@ class parpool(object):
 
     @args.setter
     def args(self, args):
-        argss = dumps(args, recurse=True)
+        argss = dill.dumps(args, recurse=True)
         self._args = (args, hash(argss), argss)
 
     @property
@@ -189,7 +266,7 @@ class parpool(object):
 
     @kwargs.setter
     def kwargs(self, kwargs):
-        kwargss = dumps(kwargs, recurse=True)
+        kwargss = dill.dumps(kwargs, recurse=True)
         self._kwargs = (kwargs, hash(kwargss), kwargss)
 
     def __enter__(self, *args, **kwargs):
@@ -210,7 +287,7 @@ class parpool(object):
                     'debuggable error. If it doesn\'t, it\'s an error specific to parallel execution.'
                         .format(r[1]))
                 warnings.formatwarning = pfmt
-                fun, args, kwargs = [loads(f[1]) for f in r[2][1:]]
+                fun, args, kwargs = [dill.loads(f[1]) for f in r[2][1:]]
                 r = (False, r[1], fun(r[2][0], *args, **kwargs))
             self.res[r[1]] = r[2]
             if not self.bar is None:
@@ -325,7 +402,7 @@ class parpool(object):
                 hs, objs = zip(*self.cache)
                 if h in hs:
                     return objs[hs.index(h)]
-            obj = loads(ser)
+            obj = dill.loads(ser)
             self.cache.append((h, obj))
             while len(self.cache) > self.cachesize:
                 self.cache.pop(0)
@@ -339,13 +416,13 @@ def pmap(fun, iterable=None, args=None, kwargs=None, length=None, desc=None, bar
         iterable: iterable from which an item is given to fun as a first argument
         args:   tuple with other unnamed arguments to fun
         kwargs: dict with other named arguments to fun
+        length: give the length of the iterator in cases where len(iterator) results in an error
         desc:   string with description of the progress bar
         bar:    bool enable progress bar
         pbar:   bool enable buffer indicator bar
         nP:     number of workers, default: number of cpu's/3
         serial: switch to serial if number of tasks less than serial, default: 4
-
-        wp@tl20200204
+        debug:  if an error occurs in an iteration, return the erorr instead of retrying in the main process
     """
     args = args or ()
     kwargs = kwargs or {}
